%%% task declaration:
%%% task(<number>,<level>)
    %%% plan(<Diameter>,<Length>); Diameter= no. of tasks at a single level, Length= no. of levels
:- dynamic task/2.
:- dynamic count/1.
% Three tasks at level 1
task(1,1).
task(2,1).
task(3,1).

% Three tasks at level 2
task(1,2).
task(2,2).
task(3,2).

% Three tasks at level 3
task(1,3).
task(2,3).
task(3,3).

% Three tasks at level 4
task(1,4).
task(2,4).
task(3,4).

naive_plan :- task(X,1), task(Y,2),task(Z,3),task(U,4),
              write('plan: '), write(X), write(Y),write(Z),write(U),nl, fail.
naive_plan.

%%% Usage:
%%% 1). Load file task.pl to prolog:    ?- [task].
%%% 2). Initialize the search-space:    ?- make(7,8). We have 8 levels, each of 7 tasks.
%%% 3). Run the generation of plans:    ?- run(3,5).  or
%%% 4). Run with measuring time,etc:    ?- time(run(3,5)). 

plan(D,1,Plan) :- task(T,1), T =< D, Plan = [T].
plan(D,N,Plan) :- N > 1, task(T,N), T =< D, N1 is N-1, plan(D,N1,PlanN1), Plan = [T|PlanN1].

run(D,N) :- retractall(count(_)), assert(count(0)),
    plan(D,N,Plan), write(Plan), 
    retract(count(C)), C1 is C + 1, assert(count(C1)), nl, fail.
run(_,_):- nl, write('No. of plans: '), count(C),write(C).


make(D,N) :- retractall(task(_,_)), 
    makelist(D,LD), makelist(N,LN),
    member(I,LD), member(J,LN),
    assert(task(I,J)), fail.
make(_,_).

makelist(1,[1]) :- !.
makelist(N,L) :- N > 1, N1 is N-1, makelist(N1, L1), append(L1,[N],L).
