/* PLAN - program znajdowania trasy polaczen miedzy miastami */
%%% Przeszukiwanie wgłąb z nawrotami
:- dynamic(p/1).
:- dynamic(p/6).
:- dynamic(q/2).
%%% p(<state>,<decision>,<goal>,<robustnes_factor>,<min_cost>,<max_cost>). %%%
%%% p(<current_node>,<next_node>,<goal_node>,<number_of_plans>,<min_cost>,<max_cost>).%%%
d(s,v1,5).
d(s,v2,7).
d(v1,v3,5).
d(v1,v4,4).
d(v2,v4,3).
d(v2,v5,6).
d(v3,v6,4).
d(v4,v6,3).
d(v4,v7,2).
d(v5,v7,7).
d(v6,g,2).
d(v7,g,5).

linkd(X,Y,D):- d(X,Y,_,D,_,1).
linkd(X,Y,D):- d(X,Y,_,D,_,0).
linkd(X,Y,D):- d(Y,X,_,D,_,0).
%%% definition of given nodes (without goal)
%nodes([s,n1,n2,n3,n4,n5,n6]).
nodes([s,v1,v2,v3,v4,v5,v6,v7]).

%%% domknięcie symetryczne relacji d/2 
     link(X,Y,D):- d(X,Y,D).
     link(X,Y,D):- d(Y,X,D).
          
%%% główna pętla szukania trasy
%%% szukaj_trasy(Start,Cel,Robocza,Trasa,DlugoscRoboczej,DlugoscTrasy,Ograniczenie).

szukaj_trasy(Miasto,Miasto,Trasa,Trasa,Dystans,Dystans,_):- !.
szukaj_trasy(Miasto,Cel,Robocza,Trasa,Dlugosc,Dystans,Ograniczenie):-
     link(Miasto,Kandydat,D),
     not(member(Kandydat,Robocza)),
     NowaDlugosc is Dlugosc + D,
     NowaDlugosc < Ograniczenie,
     szukaj_trasy(Kandydat,Cel,[Kandydat|Robocza],Trasa,
                  NowaDlugosc,Dystans,Ograniczenie).

%%% Inicjacja poszukiwan

plan(Miasto,Next,Cel,Trasa,Dystans,Ograniczenie):-
     link(Miasto, Next,D),
     szukaj_trasy(Next,Cel,[Next,Miasto],Trasa,D,Dystans,Ograniczenie). %%% !.
                                                  
%%% Find all plans from 's' to 'g'.
     fap(Plans) :- findall(Trasa, plan(s,g,Trasa,_,100),Plans).
%%% Find all plans, lengths, count, min, max: from X through Y to G. R - restriction of max distance

     fa(X,Next,G,Plans,R):- findall(s(Plan,Dist),plan(X,Next,G,Plan,Dist,R),Plans).

%%% Generate a decision table for a given node X to goal G with  length restricted to R
     
    gen_tab(X,G,R):-  link(X,Next,_),         %%% find Next node
                   %   write(Next),nl,
                      fa(X,Next,G,Plans,R),   %%% Plans is a list [s(plan1,lenth1),...,s(plank,lenghtk)]
                      build_tab(X,Next,G,Plans),       %%% assert plans info to a table
                      fail.                   %%% force search over all Next nodes
    gen_tab(_,_,_):-    write('*** no more posiibilities ***'),nl.

%%% build_tab --- building a decision table row-by-row
    build_tab(_,_,_,[]):- !.
    build_tab(X,N,G,[s(P,L)|Other_Plans]):- 
       find_min_plan_length([s(P,L)|Other_Plans],L_min),
       find_max_plan_length([s(P,L)|Other_Plans],L_max),
       length([s(P,L)|Other_Plans],No_of_Plans),  % write(No_of_Plans),nl,
       assert(p(X,N,G,No_of_Plans,L_min,L_max)).


%%%%% To do:
%%%%% Simple: for any n \in N find all decions/nexts, routes to goal, and assert decision, no_of_plans, min and max.
%%%%% Advanced: recurion on nodes. for a node find all routes to goal (nowe drogi), but if
%%%%% a route hits a visited node, just recall all the calculated roads (excluding backtrack).
%%%%% Jedno-dukierunkowosc
%%%%% Czyszcenie bazy
%%%%% Import z mapy

%%%%% A* lub inne (Nowak) do liczenia tras(y) odniesienia.

%%% Find all plans and assert; p(<plan>).
     fap-assert:- plan(s,g,Plan), assert(p(Plan)), fail.
     fap-assert.

%%% Count: zliczanie krotnosci wezla w planach.
     count(N,_):- p(P), member(N,P), add(N), fail.
     count(N,K):- q(N,K).

add(N):- \+ q(N,_),!, assert(q(N,1)).
add(N):- retract(q(N,K)), K1 is K+1, assert(q(N,K1)).

     countall:- nodes(LN), select(N,LN,_), count(N,K), write('node: '), write(N), write('  in: '), write(K),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Utilities

writelist([]):- write('*****end*****'),nl,!.
writelist([H|T]):- write(H), nl, writelist(T).

find_min_plan_length([s(_,L)],L):-!.
find_min_plan_length([s(_,L)|Rest],L):- find_min_plan_length(Rest,LR), L =< LR,!.
find_min_plan_length([s(_,L)|Rest],LR):- find_min_plan_length(Rest,LR), L > LR.

find_max_plan_length([s(_,L)],L):-!.
find_max_plan_length([s(_,L)|Rest],L):- find_max_plan_length(Rest,LR), L >= LR,!.
find_max_plan_length([s(_,L)|Rest],LR):- find_max_plan_length(Rest,LR), L < LR.

clear:- retractall(p(_,_,_,_,_,_)).
show:-
       write('  X    N    G    NP    Min    Max '), nl,
       write('========================'), nl,
        p(X,N,G,NP,Min,Max),
       write(X), write(' | '),write(N), write(' | '), write(G), write(' | '), 
       write(NP), write(' | '),write(Min), write(' | '), write(Max), nl,
     %  write('------------------------'),nl,
       fail.

show:- nl, nl, nl.
