

/*** BAZA WIEDZY ***/
fact(n(a,b)).
fact(n(a,c)).
fact(n(b,d)). 
fact(n(b,a)).
fact(n(c,a)).
fact(n(d,c)).
fact(n(d,b)).
fact(n(c,d)).

/*** STAN POCZATKOWY ***/
state([at(robot,d),at(box,a)]).

/*** STAN DOCELOWY ***/
/* end([at(robot,c),at(box,b)]). */
end([at(robot,a),at(box,d)]).

/*** OPERACJE ***/

action(go(X,Y),
     [at(robot,X),n(X,Y)],
     [at(robot,X)],
     [at(robot,Y)]).
action(move(X,Y),
     [at(box,X),n(X,Y),at(robot,X)],
     [at(robot,X),at(box,X)],
     [at(robot,Y),at(box,Y)]).



/*** INTERPRETER ***/
run :-
   initialize(Statelist,Actionlist),
   solve(Statelist,Actionlist,Result),
   stop(Result).
   
initialize(Statelist,Actionlist) :-
   state(State),
   Statelist  = [State],
   Actionlist = [],
   asserta(node(1)).
   

%%% Man Loop - Search for Plan
solve(Statelist,Actionlist,Result) :-
   end(Endlist),
   Statelist=[State|_],
   inlist(State,Endlist),
   !,
   Actionlist=Result.

solve(Statelist,Actionlist,Result) :-
   action(Action,Prec,Del,Add),
   Statelist=[State|_],   
   check(Prec,State),
   delete(State,Del,DelState),
   add(DelState,Add,NewState),
   not(instatelist(NewState,Statelist)),
   modify,
   write(NewState), nl,
   solve([NewState|Statelist],[Action|Actionlist],Result).

modify :- retract(node(N)),!,N1 is N+1,asserta(node(N1)).

count([],0) :- !.
count([_|L],N) :-
   count(L,NL), N is NL+1.
   
stop(Result) :-
   nl,nl,nl,
   write('***** PLAN *****'),
   nl,
   writeback(Result),
   count(Result,N),
   node(M),
   write('P='),write(N),write('/'),write(M),
   nl.

writeback([]) :- !, nl.
writeback([F]) :- !, write(F),nl.
writeback([F|L]) :- writeback(L), writeback([F]).
     





/*** PREDYKATY POMOCNICZE ***/

in(X,[X|_]) :- !.          /* mozna usunac determinizm */
in(X,[_|L]) :- in(X,L).

inlist([],_) :- !.
inlist([X|L],T) :-
   in(X,T),
   inlist(L,T).

instatelist(S,[P|_]) :-
   inlist(S,P),!.
instatelist(S,[_|L]) :-
   instatelist(S,L).      

check([],_) :- !.
check([F|L],State) :-
   in(F,State),
   !,
   check(L,State).
check([F|L],State) :-
   fact(F),
   check(L,State).

      
delete(State,[],State) :- !.
delete(State,[S|L],DelState) :-
   delfact(State,S,DState),
   delete(DState,L,DelState).

delfact(List,F,List) :- not(in(F,List)), !.
delfact([F|L],F,List) :- !, delfact(L,F,List).
delfact([F|L],Q,List) :- delfact(L,Q,LL), List =[F|LL].

add(State,[],State) :- !.
add(State,[F|L],NewState) :-
   addfact(State,F,FState),
   add(FState,L,NewState).

addfact(State,F,NewState) :-
   not(in(F,State)),
   !,
   NewState=[F|State].
addfact(State,_,State).   
   
   
 

