

/*** BAZA WIEDZY ***/


/*** STAN POCZATKOWY ***/
state([onfloor(a),onfloor(b),on(c,a),cl(c),cl(b)]).

/*** STAN DOCELOWY ***/

end([on(a,b),on(b,c)]).

/*** OPERACJE ***/

action(put(X,Y),
     [onfloor(X),cl(X),cl(Y)],
     [onfloor(X),cl(Y)],
     [on(X,Y)]).
action(move(X,Y,Z),
     [on(X,Y),cl(X),cl(Z)],
     [on(X,Y),cl(Z)],
     [on(X,Z),cl(Y)]).
action(free(X,Y),
     [on(X,Y),cl(X)],
     [on(X,Y)],
     [onfloor(X),cl(Y)]).


/*** INTERPRETER ***/
run :-
   initialize(Statelist,Actionlist),
   solve(Statelist,Actionlist,Result),
   stop(Result).
   
initialize(Statelist,Actionlist) :-
   state(State),
   Statelist  = [State],
   Actionlist = [],
   asserta(node(1)).
   

%%% Man Loop - Search for Plan
solve(Statelist,Actionlist,Result) :-
   end(Endlist),
   Statelist=[State|_],
   inlist(Endlist,State),
   !,
   Actionlist=Result.

solve(Statelist,Actionlist,Result) :-
   action(Action,Prec,Del,Add),
   Statelist=[State|_],   
   check(Prec,State),
   write(Prec), write('  888  '),write(State),nl,
   delete(State,Del,DelState),
   write(DelState),nl,
   add(DelState,Add,NewState),
write(NewState),nl,
   not(instatelist(NewState,Statelist)),
write(' *** '),
   not(imposs(NewState)),
   modify,
   write(NewState), nl,write(Action),nl,nl,
   solve([NewState|Statelist],[Action|Actionlist],Result).

modify :- retract(node(N)),!,N1 is N+1,asserta(node(N1)).

count([],0) :- !.
count([_|L],N) :-
   count(L,NL), N is NL+1.
   
stop(Result) :-
   nl,nl,nl,
   write('***** PLAN *****'),
   nl,
   writeback(Result),
   count(Result,N),
   node(M),
   write('P='),write(N),write('/'),write(M),
   nl.

writeback([]) :- !, nl.
writeback([F]) :- !, write(F),nl.
writeback([F|L]) :- writeback(L), writeback([F]).
     





/*** PREDYKATY POMOCNICZE ***/

in(X,[X|_]).          /* mozna usunac determinizm */
in(X,[_|L]) :- in(X,L).

inlist([],_).
inlist([X|L],T) :-
   in(X,T),
   inlist(L,T).

instatelist(S,[P|_]) :-
   inlist(S,P).
instatelist(S,[_|L]) :-
   instatelist(S,L).      

check([],_).
check([F|L],State) :-
   in(F,State),
   check(L,State).
check([F|L],State) :-
   fact(F),
   check(L,State).

imposs(State):-
   inlist([on(X,X)],State).

      
delete(State,[],State) :- !.
delete(State,[S|L],DelState) :-
   delfact(State,S,DState),
   delete(DState,L,DelState).

delfact(List,F,List) :- not(in(F,List)), !.
delfact([F|L],F,List) :- !, delfact(L,F,List).
delfact([F|L],Q,List) :- delfact(L,Q,LL), List =[F|LL].

add(State,[],State) :- !.
add(State,[F|L],NewState) :-
   addfact(State,F,FState),
   add(FState,L,NewState).

addfact(State,F,NewState) :-
   not(in(F,State)),
   !,
   NewState=[F|State].
addfact(State,_,State).   
   
   
 

